class NestRequest:
    def __init__(self, request):
        self.request = request
        self.params = NestRequestParams(self.request)
        self.query = {}
        self.body = request.get_json(silent=True, force=True)
        self.headers = {}

        if self.body is None:
            self.body = {}


class NestRequestParams(object):
    def __init__(self, request):
        super(NestRequestParams, self).__init__()
        self.request = request

    def __getattribute__(self, __name: str):
        if __name == 'request':
            return super(NestRequestParams, self).__getattribute__(__name)
        else:
            return self.request.view_args[__name]
            # return self.request.args.get(__name, type=str)



from bottlenest.core.NestContainer import NestContainer
from bottlenest.core.NestLogger import NestLogger
from bottlenest.transports.http.HttpTransport import HttpTransport


class NestApplicationContext:
    def __init__(self, module, transport):
        self.module = module
        self.logger = self.setupLogger()
        self.container = self.setupContainer()
        self.logger.setContext(self)
        self.transport = self.setupTransport(transport)
        # self.logger.log('NestApplicationContext initialized')
        # self.logger.log('NestApplicationContext module: ' +
        #                 str(self.module.name))
        # self.logger.log('NestApplicationContext options: ' + str(self.options))
        # self.logger.log('NestApplicationContext container: ' +
        #                str(self.container))
        # self.logger.log('NestApplicationContext logger: ' + str(self.logger))
        # ? should this setup be here?
        self.setup()

    def setup(self):
        self.app = self.transport
        # self.container.set('app', self.app)
        self.container.set('transport', self.transport)
        self.container.set('NestApplicationContext', self)
        self.container.set('NestContainer', self.container)
        self.module.setup(self.container)

    def setupLogger(self):
        logger = NestLogger()
        return logger

    def setupContainer(self):
        container = NestContainer()
        container.set('module', self.module)
        container.set('logger', self.logger)
        return container

    def setupTransport(self, transport):
        if transport is None:
            transport = HttpTransport()
        transport.setup(context=self.container)
        return transport

    def listen(self):
        self.logger.log('NestApplicationContext listen')

        def callback():
            self.logger.log(f"NestApplicationContext listening")
        self.transport.listen(callback)
        return self.app



from abc import ABC, abstractmethod, ABCMeta, abstractproperty
from bottlenest.metaClasses.NestProviderContext import NestProviderContext


class NestProvider(ABC):
    # def __call__(cls, *args, **kwargs):
    #    instance = super(NestProvider, cls).__call__(*args, **kwargs)
    #    # instance.setup(*args, **kwargs)
    #    return instance

    # @abstractmethod
    @property
    def name(self):
        return self.cls.__name__

    @property
    @abstractmethod
    def eventName(self):
        return 'NestEvent'

    def __init__(self, cls):
        self.cls = cls
        # self.name = cls.__name__

    def _getEventNames(self):
        eventClassName = self.eventName()
        eventNames = dir(self.provider)
        eventNames = [name for name in eventNames if type(
            getattr(self.provider, name)).__name__ == eventClassName]
        return eventNames

    def setup(self, module, context):
        self._setup(module, context)

    def _setup(self, module, context):
        print("thiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiis setup")
        self.module = module
        self.context = context
        eventContext = NestProviderContext(self)
        self.provider = self.cls(eventContext)
        eventNames = self._getEventNames()
        for eventName in eventNames:
            # print("---->> eventName: ", eventName)
            event = getattr(self.provider, eventName)
            event.setup(self.provider, context)



from flask import Flask
from bottlenest.transports.http.errors import HttpError
import traceback


class HttpTransport:
    def __init__(self, port=3500):
        self.port = port

    def setup(self, context):
        self.context = context
        self.module = self.context.get('module')
        self.logger = self.context.get('logger')
        self.app = Flask(self.module.name)
        self.context.set('app', self.app)
        self.setupErrorHandlers()

    def setupErrorHandlers(self):
        @self.app.errorhandler(Exception)
        def defaultErrorHandler(e):
            self.logger.log(
                'NestApplicationContext handle_exception', e, traceback.format_exc())
            # check if e is an instance of HttpError
            if isinstance(e, HttpError):
                return e.toDict(), e.statusCode
            return {"messages": [e.__str__()]}, 500

    def listen(self, callback):
        self.app.run(port=self.port, debug=True)
        callback()



def Injectable():
    def wrapper(providerClass):
        return NestInjectable(
            providerClass=providerClass,
        )
    return wrapper


class NestInjectable:
    def __init__(self, providerClass):
        self.name = providerClass.__name__
        self.__name__ = providerClass.__name__
        self.providerClass = providerClass
        self.instance = None
        # self.dependencies = []

    def setup(self, module, container):
        # self.dependencies = []
        # for dep in self.dependencies:
        #     dep.setup(container)
        self.module = module
        self.container = container
        container.set(self.getProviderName(), self)
        self.instance = self.providerClass(container)

    def getProviderName(self):
        providerName = f"{self.module.name}.{self.name}"
        return providerName

    def __repr__(self):
        return f"{self.name}()"

    def __str__(self):
        return f"{self.name}()"



from flask import request
from functools import wraps
from bottlenest.transports.http.decorators.NestRequest import NestRequest


class NestRoute:
    __name__ = 'NestRoute'

    def __init__(self, path, method, callback):
        print(f"route defined {method} {path}")
        self.callback = callback
        self.path = self._nestjsToFlaskPath(path)
        self.method = method

    def setup(self, cls, context):
        print(f"init route {self.path}")
        app = context.get('app')
        app.add_url_rule(
            self.path,
            methods=[self.method],
            view_func=self._callbackWrapper(cls),
        )

    def _callbackWrapper(self, cls):
        @wraps(self.callback)
        def wrapped(*args, **kwargs):
            return self.callback(cls, NestRequest(request))
            # return self.callback(context)
        return wrapped

    # Flask routing is different than NestJS routing
    def _nestjsToFlaskPath(self, path: str) -> str:
        result = ''
        parts = path.split('/')
        for part in parts:
            if part.startswith(':'):
                result += '/<' + part[1:] + '>'
            else:
                result += '/' + part
        return result[1:]



import eventlet
import socketio
from bottlenest.metaClasses.NestProvider import NestProvider
from bottlenest.transports.websockets.decorators.NestSubscribeMessage import NestSubscribeMessage


class NestWebSocketGateway(NestProvider):
    __name__ = 'NestWebSocketGateway'

    def __init__(self, cls, port=4001, namespace=None):
        self.cls = cls
        self.port = port
        self.namespace = namespace

    def eventName(self):
        return NestSubscribeMessage.__name__

    def setup(self, module, context):
        self.module = module
        self.context = context
        sio = context.get('sio')
        context.set('port', self.port)

        def _onConnect(sid, environ, auth):
            print(f"[NestWebsocketGateway] onConnect {sid}")

        sio.on('connect')(_onConnect)

        self._setup(module, context)



from abc import ABC, abstractmethod, ABCMeta

# this context is given to the controller


class NestProviderContext(ABC):
    def __init__(self, cls):
        self.provider = cls

    def get(self, key):
        getName = f"{self.provider.module.name}.{key}"
        return self.provider.context.get(getName)

    def inject(self, injectable):
        key = injectable.__name__
        getName = f"{self.provider.module.name}.{key}"
        provider = self.provider.context.get(getName)
        if provider is None:
            raise Exception(f"Provider not found: {getName}")
        return provider.instance



from bottlenest.transports.http.errors.HttpError import HttpError


class BadRequestError(HttpError):
    def __init__(self, messages=[]):
        super(BadRequestError, self).__init__(messages, 400)



class NestHttpModule(object):
    def __init__(self, moduleClass, imports, controllers, providers):
        self.name = moduleClass.__name__
        self.imports = imports
        self.controllers = controllers
        # self.moduleClass = moduleClass
        # self.moduleInstance = moduleClass()
        self.name = moduleClass.__name__
        print("moduleClass", moduleClass)
        self.providers = providers
        self.module = moduleClass()
        print(f"NestModule init", [self.name])

    def setup(self, context):
        for module in self.imports:
            module.setup(context)
        for provider in self.providers:
            provider.setup(self, context)
        for controller in self.controllers:
            controller.setup(self, context)

    # proxy static methods
    def __getattr__(self, name):
        return getattr(self.module, name)



class NestContainer:
    def __init__(self):
        self.container = {}

    def set(self, key, value):
        print("NestContainer set", key, value)
        self.container[key] = value

    def get(self, key):
        print("NestContainer get", key)
        return self.container[key] if key in self.container else None

    def inject(self, injectable):
        moduleName = self.container['module'].name
        providerName = injectable.__name__
        key = f"{moduleName}.{providerName}"
        provider = self.container[key] if key in self.container else None
        if provider:
            return provider.instance
        else:
            raise Exception(f"Provider not found: {key}")
